#!/bin/sh
#
# Script to build production distribution of mpgedit
# 
# 1) Build directory for release build:
#    mkdir mpgedit_0.6
#    mkdir mpgedit_0.6/src
#
# 2) Fetch release code from CVS:
#    cvs co -r  mpgedit_release-0-6_beta_rc1 mpgedit
#
# 3) Perform platform builds
#
# 4) Run this script:
#    ./make_dist.sh
#

RELEASE=0.6p1

dir=../..
if [ ! -d $dir/linux ]; then
  ./man2html.sh
  (cd $dir; mkdir linux solaris win32 hpux10 aix)
  (cd $dir/linux;   s_fetchtree -l -s ../src)
  (cd $dir/solaris; s_fetchtree -l -s ../src)
  (cd $dir/win32;   s_fetchtree -l -s ../src)
  (cd $dir/hpux10;  s_fetchtree -l -s ../src)
  (cd $dir/aix;     s_fetchtree -l -s ../src)
  (cd $dir/linux/mpgedit; configure)
  echo "You must perform platform specific builds for win32 and solaris"
  echo "Afterwards, execute this script again"
  exit 0
fi

if [ ! -f $dir/linux/mpgedit/mpgedit -o \
     ! -f $dir/win32/mpgedit/mpgedit.exe ]; then
  
  echo "You must perform platform specific builds for win32 and solaris"
  echo "Afterwards, execute this script again"
  exit 0
fi

uid=`id | awk '{print $1}' | sed -e 's/uid=//' -e 's/(.*//'`
if [ $uid -ne 0 ]; then
  echo "you must be root to execute this script"
  exit 1
fi

#
# Build the distribution packages areas
#
(cd $dir/linux/mpgedit; ./dist_linux.sh)
(cd $dir/win32/mpgedit; ./dist_win32.sh)
./dist_src.sh

#
# Tar/zip them up
#
(cd $dir; tar zcf mpgedit_${RELEASE}_linux.tar.gz mpgedit_0.6_linux);
(cd $dir; tar zcf mpgedit_${RELEASE}.tar.gz       mpgedit_0.6);
(cd $dir; zip mpgedit_${RELEASE}_win32.zip mpgedit_0.6_win32/*)
