/*
 * mpgedit application code
 *
 * Copyright (C) 2001, 2002 Adam Bernstein
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307  USA.
 */


#ifndef lint
static char SccsId[] = "$Id: mp3_header.c,v 1.56.4.6 2002/05/04 09:53:08 number6 Exp $";
#endif

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <sys/stat.h>
#include "portability.h"
#include "header.h"
#include "mp3time.h"
#include "mpegindx.h"
#include "xing_header.h"
#include "mpegfio.h"
#include "mp3_header.h"
#include "mpegcurses.h"
#include "decoder.h"
#include "plugin.h"
#include "version.h"


typedef struct _cmdflags
{
    editspec *edits;
    int      edits_cnt;
    char *out_filename;
    int  out_fileappend;
    int O_flag;
    int v_flag;
    int p_flag;
    int e_flag;
    int X_flag;
    char *X_args;
    int X_val;
    int c_flag;
    int s_flag;
    int has_xing;
} cmdflags;


typedef struct _mpeg_file_stats
{
    char *filename;
    int min_bitrate;
    int max_bitrate;
    int avg_bitrate;
    int file_size;
    int num_frames;
    long bitrate_frame_cnt[16];
    int mpeg_version_index;
    int mpeg_layer;
} mpeg_file_stats;


static char *g_progname;

/* This must be named "fp_mpgdecoder" for macros in plugin.h to work */
static plugin_fcn_dsc *fp_mpgdecoder;

/* Leading plus forces POSIX processing of argument list */
#define OPTARGS "+hHf:vpe:X::o:csV"

void usage(int longhelp)
{
    fprintf(stderr, "usage: %s %s\n", 
            g_progname, "-V | -h | -H | [-c [input_file [input_file2] ... ]]");
    fprintf(stderr, "       %s %s%s%s\n", g_progname,
         "[-p] [-v[v[v]]] ",
         "[-e [begint[-[endt]]]] [-X[0|1]] -f input_file ",
         "-o [+]outfile_name ");
    if (!longhelp) {
        exit(1);
    }

    if (longhelp == 2) {
        fprintf(stderr, 
"Where\n "
"  -c [input file [input_file2] ...]\n"
"      Curses edit mode.  Enter a full screen, interactive edit session\n"
"      to playback input files and set begin and end times for each\n"
"      edit.  This mode is the default action when no arguments\n"
"      are supplied.\n\n"
"  -estart[-[end]] \n"
"      Perform copy/paste operation on input files. The input MP3\n"
"      file may be encoded as either CBR or VBR.  For quick and accurate\n"
"      random access into VBR encoded files, an external time index file\n"
"      is created.  The index file uses the basename of the input file,\n"
"      with the .idx extension.\n"
"      \n"
"      The output files created by this option are located in the\n"
"      same directory as the input MP3 file.  When the -o option is\n"
"      not used, the copy/paste data is written to a file\n"
"      with the same basename as the input file, with the '_nn.mp3'\n"
"      extension, where nn is an integer value. When a file with\n"
"      basename_nn.mp3 already exists, the integer extension is incremented\n"
"      until the filename is unique.  The input MP3 file is never\n"
"      overwritten by this option.\n"
"      \n"
"      When just a begin time is specified, the copy/paste operation is\n"
"      performed from that begin time through the end of file. When just\n"
"      an end time is specified, the copy/paste operation is performed\n"
"      from the beginning of the file through the specified end time.\n"
"      When both begin and end times are specified, the portion of the\n"
"      file bracketed by these times is copy/pasted.  Begin and end times\n"
"      are specified in seconds, or MM:SS.mmm, where MM are minutes, SS\n"
"      are seconds, and mmm are milliseconds.  Although any millisecond\n"
"      value may be specified, it will always be rounded off to the next\n"
"      frame boundary, approximately 26ms.\n"
"      \n"
"      Multiple -e options are allowed. Each -e represents a separate edit\n"
"      that will be performed on the named input file that follows the last\n"
"      -e option.\n\n"
"  -f  input file name\n"
"      Multiple -f options are allowed.  At least one -e option must be\n"
"      preceed each -f option.\n\n"
"  -o  output file\n"
"      Prefixing a + to filename causes output from edit to be appended\n"
"      to the output file.  XING headers are automatically fixed up when\n"
"      using this option.\n\n"
"  -h  help information\n\n"
"  -p  play MPEG file using external player\n\n"
"  -s  silent operation\n"
"      All of the normal informational messages are suppressed during\n"
"      indexing and editing.\n\n"
"  -v  verbose output; adding more -v increases verbosity.\n"
"      Up to -vvv is allowed. -v displays the frame count and file\n"
"      offset at one second increments. -vv adds to -v this information\n"
"      for each frame: time offset, bitrate, frame size, frame count,\n"
"      frame position.  -vvv adds to -vv and displays a detailed breakdown\n"
"      of each MPEG frame header.\n\n"
"  -V  display mpgedit version and build number\n\n"
"  -X[0|1]\n"
"      Manipulate XING header. 0 omit XING header prefix, allowing output\n"
"      to be appended to the end of a previous edit. 1 fixes up XING header\n"
"      after catenating edits. This option is largely obsolete because of '+'\n"
"      option to -o\n"
);
    }
    else {
        fprintf(stderr, "\n    For detailed mpgedit usage, use -H\n");
    }
    exit(1);
}



static void myprintf(void *ctx, const char *fmt, ...);

void mpeg_file_stats_init(mpeg_file_stats *s, char *filename, int offset)
{
    memset(s, 0, sizeof(*s));
    s->min_bitrate = 1000000;
    s->max_bitrate = -1;
    s->filename    = filename;
    s->file_size   = offset;
}


void mpeg_file_stats_gather(mpeg_file_stats *stats, mpeg_header_data *header)
{
    stats->num_frames++;
    if (header->bit_rate > stats->max_bitrate) {
        stats->max_bitrate = header->bit_rate;
    }
    if (header->bit_rate < stats->min_bitrate) {
        stats->min_bitrate = header->bit_rate;
    }
    stats->avg_bitrate       += header->bit_rate;
    stats->file_size         += header->position;
    stats->file_size         += header->frame_size;
    stats->bitrate_frame_cnt[header->br_index]++;
    stats->mpeg_version_index = header->mpeg_version_index;
    stats->mpeg_layer         = header->mpeg_layer_index;
}


static void mpeg_file_stats_print(mpeg_file_stats *stats,
                           mpeg_time *time)
{
    int  min;
    long sec;
    long seconds;
    long usec;
    int  i;
    int bitrate;
    long br_count;
    int percent;
    char percent_str[32];
    int is_vbr = 0;

    mpeg_time_gettime(time, &sec, &usec);
    seconds = sec;
    min = sec / 60;
    sec %= 60;

    printf("\n\n");
    if (stats->num_frames == 0) {
        printf("File is empty, or does not contain MPEG audio data\n");
        return;
    }

    if (stats->filename) {
        printf("File name:    %s\n", stats->filename);
    }
    if (stats->max_bitrate != stats->min_bitrate) {
        is_vbr = 1;
        printf("VBR Min:      %d\n", stats->min_bitrate);
        printf("VBR Max:      %d\n", stats->max_bitrate);
        printf("VBR Average:  %d\n", stats->avg_bitrate / stats->num_frames);
    }
    else {
        printf("CBR:          %d\n", stats->max_bitrate);
    }
    printf("Total frames: %d\n", stats->num_frames);
    printf("File size:    %d\n", stats->file_size);
    printf("Track length: %d:%.2d.%d (%lds)\n",
           (int) min, (int) sec, (int) (usec / 1000), seconds);

    if (is_vbr) {
        printf("\nVariable Bit Rate frame statistics\n");
        printf("Bit Rate      Frame Count\n");
        printf("-------------------------\n");
        for (i=0; i<16; i++) {
            br_count = stats->bitrate_frame_cnt[i];
            if (br_count > 0) {
                bitrate = 
                    get_bitrate_from_index(stats->mpeg_version_index,
                                           stats->mpeg_layer,
                                           i);
                percent = br_count * 100 / stats->num_frames;
                if (percent == 0) {
                    sprintf(percent_str, "(<1%%)");
                }
                else {
                    sprintf(percent_str, "(%d%%)", percent);
                }
                printf("%3d kbps:     %6ld %s\n",
                       bitrate, br_count, percent_str);
                    
            }
        }
    }
    printf("\n\n");
}


/*
 * Perform edit; time format=mm:ss[.mmm][-mm:ss[.mmm]]
 * Note: mm:ss can also be just ss
 * The -e option without any arguments means generate
 * an index file, useful for random access computation
 * in VBR files.
 */
static void mpeg_time_string2time(char *s, mpeg_time *tval)
{
    char *sp;
    long sec  = 0;
    long usec = 0;

    if (!s) {
        return;
    }
    sp = s = strdup(s);
    if (!s) {
        return;
    }

    sec = strtol(sp, &sp, 10);
    if (sp && *sp == ':') {
        sec *= 60;
        sec += strtol(sp+1, &sp, 10);
    }
    if (sp && *sp == '.') {
        usec = strtol(sp+1, &sp, 10) * 1000;
        if (usec > 999000) {
            usec = 999000;
        }
    }
    mpeg_time_init(tval, sec, usec);
    free(s);
}


static int init_edit_times(editspec *edits, int cnt)
{
    char *name = NULL;
    char *sp;
    char *ep;
    long sec;
    long usec;
    int  i;
    int status = 1;
    char *tmptimespec;

    for (i=cnt-1; i>=0; i--) {
        sec = usec = 0;

        if (edits[i].timespec) {
            /*
             * Multiple time specs can be specified before the file name.
             * Associate all such specifications with the filename
             * that follows.
             */
            if (edits[i].filename) {
                name = edits[i].filename;
            }
            else {
                if (name) {
                    edits[i].filename = strdup(name);
                }
            }

            /* Convert edit begin/end times into time structures */
            tmptimespec = strdup(edits[i].timespec);
            sp = tmptimespec;
            ep = strchr(sp, '-');
            if (ep) {
                *ep++ = '\0';
            }
            if (sp && *sp) {
                mpeg_time_string2time(sp, &edits[i].stime);
            }
            if (ep && *ep) {
                mpeg_time_string2time(ep, &edits[i].etime);
            }
            else {
                mpeg_time_init(&edits[i].etime, 1000000, 0);
            }
            free(tmptimespec);
        }
        else {
            mpeg_time_init(&edits[i].etime, 1000000, 0);
            edits[i].timespec = strdup("");
        }
    }
    if (!edits[0].filename) {
        status = 0;
    }
    return status;
}


/*
 * Split a path into directory and filename component.  This function
 * understands relative, full path, and Windows drive letter specifications.
 * There are four important cases to be considered:
 *      input path       returned dir returned file
 *      ----------       ------------ -------------
 *   1) file             dir=./       file=file
 *   2) /dir/file        dir=/dir     file=file
 *   3) d:file           dir=d:       file=file
 *   4) d:/dir/file      dir=dir:/dir file=file
 *
 *   Returns pointer to directory, and file, or NULL on failure.
 */
static char *parse_pathname(char *path, char **ret_filebase, char **ret_ext)
{
    char *tmp_path;
    char *cp;
    char *dir  = NULL;
    char *file = NULL;
    char *ext  = NULL;

    tmp_path = strdup(path);
    if (!tmp_path) {
        goto clean_exit;
    }

    /* Convert all \ to / in path */
    for (cp = tmp_path; *cp; cp++) {
        if (*cp == '\\') {
            *cp = '/';
        }
    }

    cp = strrchr(tmp_path, '/');
    if (cp) {
        /*
         * Cases 2 and 4 above
         */
        *cp++ = '\0';
        dir = strdup(tmp_path);
        file = strdup(cp);
    }
    else if (isalpha(tmp_path[0]) && tmp_path[1] == ':') {
        /*
         * Case 3 above
         */
        dir = calloc(1, 3);
        if (!dir) {
            goto clean_exit;
        }
        dir[0] = tmp_path[0];
        dir[1] = ':';
        file = strdup(tmp_path+2);
    }
    else {
        /*
         * Case 1 above
         */
        dir = calloc(1, 3);
        if (!dir) {
            goto clean_exit;
        }
        strcpy(dir, ".");
        file = strdup(path);
    }
    cp = strrchr(file, '.');
    if (cp) {
        *cp++ = '\0';
        ext = strdup(cp);
    }
    if (!ext) {
        ext = strdup("");
    }

clean_exit:
    if (tmp_path) {
        free(tmp_path);
    }
    if (!dir || !file) {
        if (dir) {
            free(dir);
            dir = NULL;
        }
        if (file) {
            free(file);
            file = NULL;
        }
    }
    else {
        *ret_filebase = file;
        *ret_ext      = ext;
    }
    return dir;
}


/*
 * Given an input path to an mp3 file, return the path to the corresponding
 * index file.  One rule applies: when the directory containing the file
 * is read-only, the returned index path is in the local directory.
 */
static char *build_index_filename(char *path)
{
    char *dir        = NULL;
    char *filebase   = NULL;
    char *ext        = NULL;
    char *index_file = NULL;
    int  len;

    dir = parse_pathname(path, &filebase, &ext);
    if (!dir) {
        goto clean_exit;
    }

    /*
     * Is the directory read-only?
     */
    if (access(dir, W_OK) == -1) {
        free(dir);
        dir = strdup(".");
        if (!dir) {
            goto clean_exit;
        }

        /*
         * Is the directory still read-only? If so, fatal error, because
         * cannot write to the local directory.
         */
        if (access(dir, W_OK) == -1) {
            goto clean_exit;
        }
    }

    /*
     * Build the index file name. Length is dir + / + filebase + .idx + \0
     */
    len = strlen(dir) + 1+ strlen(filebase) + 4 + 1;
    index_file = calloc(1, len);
    if (!index_file) {
        goto clean_exit;
    }
    strcpy(index_file, dir);
    if (strlen(dir) != 2 || dir[1] != ':') {
        strcat(index_file, "/");
    }
    strcat(index_file, filebase);
    strcat(index_file, ".idx");
    
clean_exit:
    if (dir) {
        free(dir);
    }
    if (filebase) {
        free(filebase);
    }
    if (ext) {
        free(ext);
    }
    return index_file;
}


static int mp3edit_file_exists(char *base, char *ext, char *offset, int n)
{
    sprintf(offset, "_%d.%s", n, ext);
    return access(base, F_OK) == 0;
}


/*
 * Find the next available output file.   Given a file with the name
 * name.mp3, this function builds the name name_NNN.mp3, where NNN
 * is an integer number.  This function finds the next unused file
 * name in the NNN sequence. 
 *
 * The search for the next available file is implemented as a binary
 * search.  The search starts by finding the first unused filename, by
 * incrementing NNN by powers of 2.  Once this high number is found,
 * then the binary search begins.
 *
 * This function modifies the space pointed at by base.  This
 * function is an internal helper routine for build_edit_filename();
 */
static void mp3edit_find_free_slot(char *base, char *ext, char *offset)
{
    int low  = 1;
    int high = 1;
    int test;
    int rsts;

    /*
     * Search for first available unused slot
     */
    rsts = mp3edit_file_exists(base, ext, offset, high);
    while (rsts) {
        low = high;
        high *= 2;
        rsts = mp3edit_file_exists(base, ext, offset, high);
    }

    /* found slot if first numbered file does not exit */
    if (low == high) {
        return;
    }

    /* Perform binary search for next available slot */
    do {
        test = low + ((high - low) / 2);
        rsts = mp3edit_file_exists(base, ext, offset, test);
        if (rsts) {
            low = test;
        }
        else {
            high = test;
        }
    } while ((low+1) != high);

    /* Format the file name using the high value */
    mp3edit_file_exists(base, ext, offset, high);
}


/*
 * Given an input path to an mp3 file, return the path to the corresponding
 * edit output file.  One rule applies: when the directory containing the file
 * is read-only, the returned output file is in the local directory.
 * This function builds a unique output file name, by adding the suffix
 * _NNN between the filebase and the file extension.  Given /path/file.mp3,
 * the output file name /path/file_1.mp3 is created.  Should that file exist,
 * the integer suffix is incremented until the file does not exist.
 */
static char *build_edit_filename(char *infile_name)
{
    char *cp;
    char *edit_filename = NULL;
    char *dir;
    char *filebase = NULL;
    char *ext      = NULL;
    int  len;

    dir = parse_pathname(infile_name, &filebase, &ext);
    if (!dir) {
        goto clean_exit;
    }

    /*
     * Is the directory read-only?
     */
    if (access(dir, W_OK) == -1) {
        free(dir);
        dir = strdup(".");
        if (!dir) {
            goto clean_exit;
        }
        /*
         * Is the directory still read-only? If so, fatal error, because
         * cannot write to the local directory.
         */
        if (access(dir, W_OK) == -1) {
            goto clean_exit;
        }
    }

    /*
     * Convert /path/prefix.mp3 -> /path/prefix_N.mp3.
     * New size is length of path + prefix + "_" (1) + max integer
     * size (11) + ".ext" + null byte (1).
     */
    len = strlen(dir) + 1 + strlen(filebase) + 1 + 11 + 1 + strlen(ext) + 1;
    edit_filename = (char *) calloc(1, len);
    if (!edit_filename) {
        fprintf(stdout, "failed to alloc edit filename\n");
        goto clean_exit;
    }
    strcpy(edit_filename, dir);
    if (strlen(dir) != 2 || dir[1] != ':') {
        strcat(edit_filename, "/");
    }
    strcat(edit_filename, filebase);
    cp = edit_filename + strlen(edit_filename);
    if (!ext) {
        ext = strdup(".mp3");
        if (!ext) {
            goto clean_exit;
        }
    }
    mp3edit_find_free_slot(edit_filename, ext, cp);

clean_exit:
    if (!dir || !filebase || !ext) {
        if (edit_filename) {
            free(edit_filename);
            edit_filename = NULL;
        }
    }

    if (dir) {
        free(dir);
    }
    if (filebase) {
        free(filebase);
    }
    if (ext) {
        free(ext);
    }
    return edit_filename;
}


static FILE *init_timeindex_file(FILE *fp, char *indx_filename,
                                 mpegfio_iocallbacks *ttyio)
{
    FILE *indxfp = NULL;
    char *dir    = NULL;
    char *file   = NULL;
    int  sts;

    if (!fp || !indx_filename) {
        return NULL;
    }

    indx_filename = build_index_filename(indx_filename);
    if (!indx_filename) {
        goto clean_exit;
 
    }

    /*
     * Test existing index file for validity.  If it is deemed
     * incorrect remove the file, so it will be created anew.
     */
    indxfp = fopen(indx_filename, "rb");
    if (indxfp) {
        sts = mp3edit_isvalid_index(fp, indxfp);
        if (sts != 1) {
            remove(indx_filename);
        }
        fclose(indxfp);
    }
    /*
     * Test for need to create index file.  Open index file for read; 
     * on error, open for write, because file does not exist.
     */
    if (access(indx_filename, F_OK) == -1) {
        indxfp = fopen(indx_filename, "wb");
        if (!indxfp) {
            goto clean_exit;
        }
        if (ttyio && ttyio->printf) {
            ttyio->printf(ttyio->ctx,
                          "Creating index file %s\n", indx_filename);
        }
        mp3edit_create_indexfile(fp, indxfp, ttyio);
        if (ttyio && ttyio->printf) {
            ttyio->printf(ttyio->ctx,
                          "\ndone.\n");
        }
        fclose(indxfp);
        rewind(fp);
    }
    indxfp = fopen(indx_filename, "rb");
    if (!indxfp) {
        goto clean_exit;
    }

clean_exit:
    if (file) {
        free(file);
    }
    if (dir) {
        free(dir);
    }
    if (indx_filename) {
        free(indx_filename);
    }
    return indxfp;
}


static void mpegfio_edit_segment(
                          FILE            *fp,
                          FILE            *outfp,
                          FILE            **playfp,
                          mpeg_time       *stime,
                          mpeg_time       *etime,
                          mpeg_file_stats *stats,
                          long            *file_position,
                          mpegfio_iocallbacks *ttyio,
                          int             verbose)
{
    int              found = 0;
    mpeg_file_iobuf  mpegiobuf;
    mpeg_header_data header;
    int              tlast = 0;
    int              tnow  = 0;
    long             sec;
    long             usec;
    long             etime_sec;
    long             etime_usec;
    int              inch;
    int              eof = 0;

    mpeg_file_iobuf_clear(&mpegiobuf);
    memset(&header, 0, sizeof(header));

    if (*playfp) {
        fp_mpgdecoder_init(ttyio->mpctx);
    }

    mpeg_time_gettime(etime, &etime_sec, &etime_usec);
    do {
        found = read_next_mpeg_frame(fp, &mpegiobuf, &header, &eof);
        if (found == 1) {
            mpeg_time_frame_increment(stime, &header);
            mpeg_file_stats_gather(stats, &header);
            tnow = mpeg_time_2seconds(stime);
            *file_position += header.position;
            if (ttyio && ttyio->printf && verbose > 2) {
                ttyio->printf(ttyio->ctx, "MPEG header ");
            }
            mpeg_time_gettime(stime, &sec, &usec);
            if (ttyio && ttyio->printf && verbose > 1) {
                ttyio->printf(ttyio->ctx,
                "t=%d.%-.3ds  br=%-3d  sz=%-4d fr=%-6d pos=%-10ld pos=0x%lx\n",
                       (int) sec,
                       (int) (usec/1000),
                       header.bit_rate,
                       header.frame_size,
                       stats->num_frames,
                       *file_position,
                       *file_position);
            }
            if (ttyio && ttyio->printf && verbose > 2) {
/* TBD/ttyio passed in here */
                print_mpeg_header_values(&header);
            }
            *file_position += header.frame_size;

            /*
             * Write edit to file, or to external MP3 player.
             */
            if (*playfp) {
                fp_mpgdecoder_play_frame(ttyio->mpctx, playfp, 
                                      mpeg_file_iobuf_getptr(&mpegiobuf),
                                      header.frame_size, header.sample_rate,
                                      header.channel_mode != 3);
            }
            if (outfp) {
                fwrite(mpeg_file_iobuf_getptr(&mpegiobuf),
                       header.frame_size, 1, outfp);
            }

            if (sec > etime_sec || (sec == etime_sec && usec >= etime_usec)) {
                eof = 1;
            }
            if (tnow > tlast || eof) {
                tlast = tnow;
                if (ttyio && ttyio->printf && verbose == 0) {
                    ttyio->printf(ttyio->ctx, 
                        "\rElapsed time: %4d:%02d| %5d.%03ds|Frame: %-7d\r",
                           tnow/60,
                           tnow%60,
                           tnow,
                           usec/1000,
                           stats->num_frames);
                }

                if (ttyio && ttyio->printf && verbose > 0) {
                    ttyio->printf(ttyio->ctx, 
                           "  *** t=%ds fr=%d pos=%ld (0x%lx)\n",
                           tnow, stats->num_frames,
                           *file_position, *file_position);
                }

            }
            mpeg_file_iobuf_setstart(&mpegiobuf,
                                    mpeg_file_iobuf_getstart(&mpegiobuf) +
                                         header.frame_size);
        }
        /* found */

        if (*playfp && ttyio && ttyio->getch &&
            (inch = ttyio->getch(NULL)) && inch != -1)
        {
            /*
             * Pause playback if <Enter> is pressed.
             * Resume if <Enter> is pressed again, otherwise stop
             */
            if (inch == '\n') {
                fp_mpgdecoder_close(ttyio->mpctx);
                *playfp = NULL;
                do {
                    sleep(1);
                } while ((inch = ttyio->getch(NULL)) && inch == -1);
            }
            
            if (inch != '\n') {
                eof = 1;
                /*
                 * Eat any remaining input characters. This happens 
                 * when a multi-character key, like an arrow key, is pressed.
                 */
                do {
                    inch = ttyio->getch(NULL);
                } while (inch && inch != -1);
            }
            while (!eof && !*playfp) {
                *playfp = fp_mpgdecoder_open(ttyio->mpctx, header.sample_rate,
                                             header.channel_mode != 3);
            }
        }
    } while (!eof);
}


static int mpegfio_play_segment(
                         char *filename,
                         mpeg_time *stime,
                         mpeg_time *etime,
                         mpegfio_iocallbacks *ttyio)
{
    char             *indx_filename = NULL;
    int              has_xing;
    FILE             *fp = NULL;
    FILE             *playfp = NULL;
    FILE             *indxfp = NULL;
    mpeg_header_data header;
    mpeg_time        mpegtval;
    mpeg_file_stats  stats;
    long             file_position = 0;
    long             offset = 0;
    int              status = 0;
    XHEADDATA        xingh;
    char             *dir = NULL;

    indx_filename = build_index_filename(filename);
    if (!indx_filename) {
        status = 1;
        goto clean_exit;
    }

    fp = fopen(filename, "rb");
    if (!fp) {
        status = 2;
        goto clean_exit;
    }
    indxfp = fopen(indx_filename, "rb");
    if (!indxfp) {
        goto clean_exit;
    }

    memset(&header, 0, sizeof(header));
    mpeg_time_clear(&mpegtval);
    mpeg_file_stats_init(&stats, filename, 0);

    has_xing = mpegfio_has_xing_header(fp, &header, &xingh, 0);

    playfp = fp_mpgdecoder_open(ttyio->mpctx, header.sample_rate,
                                header.channel_mode != 3);
    if (!playfp) {
        status = 3;
        goto clean_exit;
    }

    offset = mpeg_time_seek_starttime(fp, indxfp, stime,
                                      &mpegtval, has_xing);
    if (offset == -1) {
        status = 5;
        goto clean_exit;
    }
    file_position += offset;
    mpegfio_edit_segment(fp,
                         NULL,
                         &playfp,
                         &mpegtval,
                         etime,
                         &stats,
                         &file_position,
                         ttyio,
                         0);


clean_exit:
    if (fp) {
        fclose(fp);
    }
    if (indxfp) {
        fclose(indxfp);
    }
    if (playfp) {
        fp_mpgdecoder_close(ttyio->mpctx);
    }
    if (indx_filename) {
        free(indx_filename);
    }
    if (dir) {
        free(dir);
    }
    return status;
}


int curs_interactive_play_file(char *filename,
                               char *stime_str,
                               char *dtime_str,
                               mpegfio_iocallbacks *ttyio)
{
    mpeg_time stime; /* Start time */
    mpeg_time etime; /* End time */
    mpeg_time dtime; /* Play duration */
    long sec;
    long usec;
    long dsec;
    long dusec;
    int status = 0;

    mpeg_time_string2time(stime_str, &stime);
    mpeg_time_gettime(&stime, &sec,  &usec);

    mpeg_time_string2time(dtime_str, &dtime);
    mpeg_time_gettime(&dtime, &dsec, &dusec);

    mpeg_time_init(&etime, sec+dsec, usec+dusec);
    status = mpegfio_play_segment(filename, &stime, &etime, ttyio);
    return status;
}


int curs_build_indexfile(char *filename, mpegfio_iocallbacks *ttyio)
{
    char *indx_filename;
    FILE *fp;
    FILE *indxfp;
    struct stat sbuf;

    if (stat(filename, &sbuf) == -1) {
        return 1;
    }

    if (!S_ISREG(sbuf.st_mode)) {
        return 1;
    }

    fp = fopen(filename, "rb");
    if (!fp) {
        return 1;
    }

    indx_filename = build_index_filename(filename);
    if (!indx_filename) {
        return 1;
    }

    indxfp = init_timeindex_file(fp, indx_filename, ttyio);
    if (!indxfp) {
        return 1;
    }
    fclose(indxfp);
    fclose(fp);
    free(indx_filename);
    return 0;
}


int curs_get_size_from_index(char *filename, long *sec, long *usec)
{
    FILE *indxfp;
    char *indx_filename;
    struct stat sbuf;
    long fsize = 0;

    if (stat(filename, &sbuf) == -1) {
        return 1;
    }

    if (!S_ISREG(sbuf.st_mode)) {
        return 1;
    }


    indx_filename = build_index_filename(filename);
    if (!indx_filename) {
        return 1;
    }
    indxfp = fopen(indx_filename, "r");
    if (!indxfp) {
        return 1;
    }

    fsize = mp3edit_get_size_from_index(indxfp, sec, usec);
    fclose(indxfp);
    free(indx_filename);
    return !(fsize > 0);
}


/*
 * Returns 0 if named file is ok
 *         1 if file already exists
 *         2 if append has been specified, and file is not writable
 */
int validate_outfile(char *filename, char **ret_filename, int *ret_fileappend) 
{
    char *edit_filename = NULL;
    int out_fileappend  = 0;
    int status          = 0;

    if (filename) {
        if (filename[0] == '+') {
            edit_filename = &filename[1];
            out_fileappend = 1;
        }
        else {
            edit_filename = filename;
        }

        if (out_fileappend) {
            if (access(edit_filename, F_OK) == 0) {
                if (access(edit_filename, W_OK) != 0) {
                    status = 2;
                    goto clean_exit;
                }
            }
        }
        else {
            if (access(edit_filename, F_OK) == 0) {
                status = 1;
                goto clean_exit;
            }
        }
    }

clean_exit:
    if (ret_filename) {
        *ret_filename = edit_filename;
    }
    if (ret_fileappend) {
        *ret_fileappend = out_fileappend;
    }
    return status;
}


static int validate_outfile_status(int status, char *edit_filename)
{
    if (status) {
        if (status == 1) {
            fprintf(stderr, "Output file '%s' already exists\n",
                    edit_filename);
        }
        else if (status == 2){
            fprintf(stderr, "Output file '%s' is not writable for append\n",
                    edit_filename);
        }
        else {
            fprintf(stderr, "Error using output file '%s'\n",
                    edit_filename);
        }
    }
    return status ? 1 : 0;
}


static int validate_edit_times(editspec *edits, int edits_cnt,
                               int silent, int *rindx)
{
    FILE                *fp = NULL;
    FILE                *indxfp = NULL;
    int                 i;
    char                *indx_filename = NULL;
    int                 rsts = 1;
    mpeg_time           mpegtval;
    int                 status;
    mpegfio_iocallbacks ttyio;
    long                spos = 0;
    long                epos = 0;

    memset(&ttyio, sizeof(ttyio), 0);
    if (silent) {
        ttyio.printf = NULL;
        ttyio.getch  = NULL;
    }
    else {
        ttyio.printf = myprintf;
        ttyio.getch  = NULL;
    }
    memset(&mpegtval, sizeof(mpegtval), 0);
    for (i=0; i<edits_cnt; i++) {
        *rindx = i;


        /*
         * Illegal to have start greater than end time
         */
        if ((edits[i].stime.units > edits[i].etime.units) ||
            (edits[i].stime.units == edits[i].etime.units && 
             edits[i].stime.usec > edits[i].etime.usec)) 
        {
            rsts = 4;
            goto clean_exit;
        }

        indx_filename = build_index_filename(edits[i].filename);
        if (!indx_filename) {
            goto clean_exit;
        }

        fp = fopen(edits[i].filename, "rb");
        if (!fp) {
            goto clean_exit;
        }

        indxfp = init_timeindex_file(fp, indx_filename, &ttyio);
        if (!indxfp) {
            rsts = 2;
            goto clean_exit;
        }

        if (fseek(fp, 0, SEEK_END) != -1) {
            epos = ftell(fp);
        }
        status = mpeg_time_seek_starttime(fp, indxfp,
                                          &edits[i].stime, &mpegtval, 0);
        if (status == -1) {
            rsts = 3;
            goto clean_exit;
        }
        spos = ftell(fp);

        /*
         * When no end time is specified, synthesize one from the start time,
         * where the end time is one frame beyond the start frame.  99.99%
         * of the time the edit will just go from the start time to EOF.
         * However, there is a case where the start time is at EOF, and this
         * edit will result in a 0 byte edit, which displays an ugly error 
         * message.
         */
        if (spos == epos) {
            rsts = 5;
            goto clean_exit;
        }
        else if (edits[i].etime.units != 1000000) {
            status = mpeg_time_seek_starttime(fp, indxfp,
                                              &edits[i].etime, &mpegtval, 0);
        }
        if (status == -1) {
            rsts = 5;
            goto clean_exit;
        }
        fclose(fp);
        fclose(indxfp);
        free(indx_filename);
        fp            = NULL;
        indxfp        = NULL;
        indx_filename = NULL;
    }
    rsts = 0;

clean_exit:
    if (fp) {
        fclose(fp);
    }
    if (indxfp) {
        fclose(indxfp);
    }
    if (indx_filename) {
        free(indx_filename);
    }
    return rsts;
}


static void myprintf(void *ctx, const char *fmt, ...)
{
    va_list ap;
    va_start(ap, fmt);
    vprintf(fmt, ap);
    fflush(stdout);
    va_end(ap);
}


/*
 * Not an exhaustive validation of the provided time specification,
 * but a consistency check.
 */
static int validate_edit_timespec(char *str)
{
    /* Null pointer or null string not valid */
    if (!str || !*str) {
        return 0;
    }

    /* 12 || :34 check */
    if (isdigit(str[0]) || str[0] == ':') {
        return 1;
    }

    /*
     * Must look at character following - to prove the value is not
     * the next option on the command line; -, -1:34 and -34 are valid.
     */
    if (str[0] == '-') {
        if (str[1] != '\0' && str[1] != ':' && !isdigit(str[1])) {
            return 0;
        }
        return 1;
    }

    return 0;
}


static char *validate_edit_times_errorstr(char *filename,
                                          char *timespec, int status)
{
    char *errstr = malloc(128+strlen(filename));

    *errstr = '\0';
    switch(status) {
      case 1:
        sprintf(errstr, "file name '%s' invalid", filename);
        break;
      case 2:
      case 3:
        sprintf(errstr, "file '%s', bad start time '%s'", filename, timespec);
        break;
      case 4:
        sprintf(errstr, "file '%s', start time is larger than end time '%s'",
                filename, timespec);
        break;
      case 5:
        sprintf(errstr, "file '%s', bad end time '%s'", filename, timespec);
        break;
      default:
        break;
    }
    return errstr;
}


void mpgedit_version(void)
{
    printf("\n%s version %s %s\n",
           MPGEDIT_PRODUCT,
           MPGEDIT_VERSION,
           MPGEDIT_URL);
    printf("Build:          %s (%s)\n", MPGEDIT_BUILDNUM, MPGEDIT_BUILDOS);
    printf("Report bugs to: %s\n", MPGEDIT_MAILTO);
    exit(1);
}


int curs_validate_edit_times(editspec *edits, int edits_cnt, int *errindx)
{
    int              status;
    int              indx;

    /*
     * Process -e and -f options.
     */
    if (!init_edit_times(edits, edits_cnt)) {
        fprintf(stderr, "%s: -f filename never specified\n", g_progname);
        return 1;
    }

    status = 0;
    if (edits_cnt > 0) {
        status = validate_edit_times(edits, edits_cnt, 0, &indx);
        if (status && errindx) {
            *errindx = indx;
        }
    }
    return status;
}


char *curs_validate_edit_times_errorstr(char *filename,
                                        char *timespec,
                                        int status)
{
    return validate_edit_times_errorstr(filename, timespec, status);
}



char *mp3edit_alloc_infile(char *name)
{
    char *filename;
    char *cp;

    filename = strdup(name);
    if (!filename) {
        return NULL;
    }

    /*
     * Make all \ directory separators a /, as much of the subdirectory
     * processing code assumes a /, and Win32 filesystem functions will
     * accept a /.
     */
    for (cp = filename; *cp; cp++) {
        if (*cp == '\\') {
            *cp = '/';
        }
    }
    return filename;
}



int main(int argc, char *argv[])
{
    FILE             *fp     = NULL;
    FILE             *playfp = NULL;
    FILE             *indxfp = NULL;
    FILE             *editfp = NULL;
    int              ch;
    long             file_position = 0;
    mpeg_time        mpegtval;
    mpeg_file_stats  stats;
    mpeg_header_data header;
    char             *indx_filename = NULL;
    char             *edit_filename = NULL;
    int              indx;
    XHEADDATA        xingh;
    mpeg_time        *stime = NULL;
    int              editsindx;
    editspec         *curs_edits;
    int              curs_edits_cnt;
    int              status;
    int              do_append = 0;
    mpegfio_iocallbacks ttyio;
    cmdflags         argvflags;
    char             *loaderrmsg;
    char             *errstr;

    g_progname = (g_progname = strrchr(argv[0], '/')) ?
                     g_progname+1 : argv[0];

    memset(&argvflags, 0, sizeof(argvflags));
    memset(&ttyio, 0, sizeof(ttyio));

    fp_mpgdecoder = mpgfio_alloc_decoder_plugin();
    if (!fp_mpgdecoder) {
        fprintf(stderr, "Failed allocating decoder plugin\n");
        return 1;
    }

    /* 
     * More memory is allocated than needed here, but this saves
     * having to pre-count the total number of -e options specified.
     */
    argvflags.edits = (editspec *) calloc(argc, sizeof(editspec));
    if (!argvflags.edits) {
        perror("calloc: cannot allocate edit list");
        return 1;
    }
    opterr = 0;
    while ((ch = getopt(argc, argv, OPTARGS)) != -1) {
        switch (ch) {
          case 'h':
            usage(1);
            break;

          case 'H':
            usage(2);
            break;

          case 'f':
            argvflags.edits[argvflags.edits_cnt].filename =
                mp3edit_alloc_infile(optarg);
            argvflags.edits_cnt++;
            break;

          case 'o':
            argvflags.out_filename = strdup(optarg);
            break;

          case 'p':
            argvflags.p_flag = 1;
            break;

          case 'v':
            argvflags.v_flag++;
            break;

          case 'e':
            argvflags.e_flag = 1;
            if (optarg) {
                if (!validate_edit_timespec(optarg)) {
                    fprintf(stderr, "%s '%s' %s\n",
                            "Error: -e argument",
                            optarg,
                            "is not a valid time specification");
                    usage(0);
                }
                if (argvflags.edits[argvflags.edits_cnt].timespec) {
                    argvflags.edits_cnt++;
                }
                argvflags.edits[argvflags.edits_cnt].timespec = strdup(optarg);
            }
            break;

          case 'X':
            argvflags.X_flag = 1;
            if (optarg) {
                argvflags.X_args = strdup(optarg);
            }
            break;

          case 's':
            argvflags.s_flag = 1;
            break;

          case 'c':
            argvflags.c_flag = 1;
            break;

          case 'V':
            mpgedit_version();
            break;

          case '?':
            fprintf(stderr, "%s: '-%c' unknown option or missing argument\n", 
                    g_progname, optopt);
            usage(0);
            break;

          default:
            fprintf(stderr, "%s: error parsing command line\n", g_progname);
            usage(0);
            break;
        }
    }
    if (argc == 1) {
        argvflags.c_flag = 1;
    }


    /*
     * It is convenient to treat the first argument following all of the
     * command line options as the input file name.  
     */
    while (optind < argc && argv[optind]) {
        /* If any option is found, that is a fatal error! */
        if (argv[optind][0] == '-') {
            fprintf(stderr, "Unexpected option found after filenames '%s'\n",
                    argv[optind]);
            fprintf(stderr, "Try using '-f' before input file names\n");
            return 1;
        }

        argvflags.edits[argvflags.edits_cnt].filename =
            mp3edit_alloc_infile(argv[optind]);

        if (!argvflags.edits[argvflags.edits_cnt].timespec) {
            argvflags.edits[argvflags.edits_cnt].timespec = strdup("");
        }
        argvflags.edits_cnt++;
        optind++;
    }

    if (argvflags.p_flag || argvflags.c_flag) {

        /*
         * Load MP3 file playback plugin.
         */
        loaderrmsg = mpegfio_load_decoder_plugin("libmpgedit_decoder",
                                                 fp_mpgdecoder);
        if (loaderrmsg) {
            fprintf(stderr,
                    "Error while loading plugin 'libmpgedit_decoder': %s\n",
                    loaderrmsg);
            fprintf(stderr, "%s\n",
                "Warning: decoder.so plugin not found; default decoder used");
            fprintf(stderr, "%s\n",
                "         set LD_LIBRARY_PATH or install with ldconfig\n");
            free(loaderrmsg);
        }
        ttyio.mpctx = fp_mpgdecoder_alloc();
        if (!ttyio.mpctx) {
            fprintf(stderr, "allocating mpgdecoder context failed\n");
            return 1;
        }
        /*
         * Open/close decoder to reveal any decoder open errors.  Also do this
         * when entering curses mode, since it is highly likely the user will
         * want to use the playback feature while in curses mode.
         */
        playfp = fp_mpgdecoder_open(ttyio.mpctx, 44100, 0);
        if (!playfp) {
            fprintf(stderr, "Unable to open playback plugin");
            return 1;
        }
        fp_mpgdecoder_close(ttyio.mpctx);
        playfp = NULL;
    }

    if (argvflags.c_flag) {
        status = curs_play(argvflags.edits, argvflags.edits_cnt,
                           &ttyio,
                           &curs_edits, &curs_edits_cnt,
                           &argvflags.O_flag,
                           &argvflags.out_filename);
        if (status) {
            return 1;
        }
        argvflags.edits_cnt = curs_edits_cnt;
        if (argvflags.edits) {
            free(argvflags.edits);
        }
        argvflags.edits = curs_edits;
        argvflags.e_flag = 1;
    }

    /* Silent operation */
    if (argvflags.s_flag) {
        argvflags.v_flag = 0;
        ttyio.printf     = NULL;
        ttyio.getch      = NULL;
    }
    else {
        ttyio.printf = myprintf;
        ttyio.getch  = NULL;
    }

    /* -o with no -e not allowed */
    if (!argvflags.e_flag && argvflags.out_filename) {
        fprintf(stderr, "%s: -o must be used with -e\n", g_progname);
        return 1;
    }

    if (!argvflags.edits_cnt) {
        fprintf(stderr, "%s: no file name specified\n", g_progname);
        return 1;
    }

    if (argvflags.out_filename &&
        (argvflags.p_flag || argvflags.X_flag))
    {
        fprintf(stderr, "%s: -o not allowed with -p or -X\n", g_progname);
        return 1;
    }

    if (argvflags.p_flag && argvflags.X_flag) {
        fprintf(stderr, "%s: both -p and -X not allowed\n", g_progname);
        return 1;
    }

    /* 
     * XING header processing. -X is same as -X0.  -X0 prevents addition
     * of XING header to edited cut. -X1 performs analysis of edited cut
     * and fixes up header.
     */
    if (argvflags.X_flag) {
        if (argvflags.X_args) {
            argvflags.X_val = atoi(argvflags.X_args);
        }
        else {
            argvflags.X_val = 0;
        }
    }

    /*
     * curs_play() has already done this initialization and
     * validation, so don't duplicate it here.
     */
    if (argvflags.c_flag == 0) {
        /*
         * Process -e and -f options.
         */
        if (!init_edit_times(argvflags.edits, argvflags.edits_cnt)) {
            fprintf(stderr, "%s: -f filename never specified\n", g_progname);
            return 1;
        }

        status = 0;
        if (argvflags.edits_cnt > 0) {
            status = validate_edit_times(argvflags.edits,
                                         argvflags.edits_cnt, argvflags.s_flag,
                                         &indx);
        }
        if (status > 0) {
            errstr = validate_edit_times_errorstr(
                         argvflags.edits[indx].filename,
                         argvflags.edits[indx].timespec, status);
            fprintf(stderr, "Error: %s\n", errstr);
            free(errstr);
            return 1;
        }
    }

    /*
     * Populate edit_filename and opt_fileappend when out_file is specified
     */
    status = validate_outfile(argvflags.out_filename,
                              &edit_filename, &argvflags.out_fileappend);
    if (status) {
        return validate_outfile_status(status, argvflags.out_filename);
    }

    /*
     * When only an input file is specified, add -v action.
     */
    if (!argvflags.v_flag && !argvflags.p_flag && !argvflags.e_flag) {
        printf("No edits specified; only reading input data\n\n");
    }

    editsindx = 0;
    do {
        fp = fopen(argvflags.edits[editsindx].filename, "rb");
        if (!fp) {
            fprintf(stderr, "%s: error opening file %s\n", 
                    g_progname, argvflags.edits[editsindx].filename);
            return 1;
        }

        indx_filename = 
            build_index_filename(argvflags.edits[editsindx].filename);
        if (!indx_filename) {
            fprintf(stdout, "failed to alloc indx filename\n");
        }

        if (argvflags.e_flag) {
            stime = &argvflags.edits[editsindx].stime;
            /*
             * Determine if the previous end time, and the current start time
             * are both zero.  This indicates that the current file edits 
             * should be appended to the previous edit_filename. Of course,
             * this is only necessary when no output file has been
             * specified, because all edits are appended in that case.
             */
           
            if (!argvflags.out_filename && editsindx > 0) {
                do_append = 
                    argvflags.edits[editsindx-1].etime.units == 1000000 &&
                    argvflags.edits[editsindx-1].etime.usec  == 0 &&
                    argvflags.edits[editsindx].stime.units == 0 &&
                    argvflags.edits[editsindx].stime.usec  == 0;
            }
            else if (argvflags.out_fileappend) {
                do_append = 1;
            }
            else {
                do_append = 0;
            }

            indxfp = init_timeindex_file(fp, indx_filename, &ttyio);
            if (!indxfp) {
                fprintf(stderr, "Error opening time index file\n");
                return 1;
            }
        }

        /*
         * Build filename edits are saved to.  When a directory path is
         * specified, output will be in the same directory as the input file.
         * An exception to this rule is when the input file directory is read
         * only.  In that case, output is written to the current directory.
         */
        if (!argvflags.out_filename) {
            if (do_append) {
                if (!editfp) {
                    editfp = fopen(edit_filename, "ab+");
                }
            }
            else {
                if (edit_filename) {
                    free(edit_filename);
                }
    
                edit_filename = build_edit_filename(
                                    argvflags.edits[editsindx].filename);
                if (!edit_filename) {
                    return 1;
                }
    
                /*
                 * Paranoia check: edit_filename should be unique at this
                 * point, as it was constructed above to be unique.
                 * Double check, since a mistake here will cause the
                 * input file data to be lost.
                 */
                status = validate_outfile(edit_filename,
                                  &edit_filename, &argvflags.out_fileappend);
                if (status) {
                    return validate_outfile_status(status, edit_filename);
                }
            }
        }

        if (!argvflags.p_flag &&
            argvflags.e_flag && argvflags.edits[editsindx].timespec)
        {
            /*
             * Open edit file when a cut begin/end time are specified.
             * Edit file is opened for read/write, so the Xing VBR header
             * can be modified.
             */
            if (do_append) {
                if (!editfp) {
                    editfp = fopen(edit_filename, "ab+");
                }
            }
            else {
                editfp = fopen(edit_filename, "wb");
            }

            if (!editfp) {
                fprintf(stdout, "failed to open edit file tmp.mp3\n");
                return 1;
            }
            if (!argvflags.s_flag) {
                printf("%s edits to file '%s'\n",
                       do_append ? "Appending" : "Writing", edit_filename);
            }
        }

        memset(&header, 0, sizeof(header));
        argvflags.has_xing =
            mpegfio_has_xing_header(
                do_append ? editfp : fp, &header, &xingh, do_append);
    
        if (argvflags.has_xing) {
            if (argvflags.v_flag > 0) {
                printf("Found xing header\n");
                xingheader_print(&xingh);
            }
    
            file_position = argvflags.has_xing;
            if (editfp && !do_append &&
                (!argvflags.X_flag ||
                (argvflags.X_flag && argvflags.X_val != 0)))
            {
                fwrite(xingh.xingbuf, argvflags.has_xing, 1, editfp);
            }
        }
        else {
            fseek(fp, 0, SEEK_SET);
        }

        mpeg_time_clear(&mpegtval);

        /*
         * argvflags.has_xing contains the size of the xing header. When
         * appending files together, do not count this size in the file
         * total, since the joined file does not have this frame.
         */
        mpeg_file_stats_init(&stats,
                             argvflags.edits[editsindx].filename,
                             do_append ? 0 : argvflags.has_xing);
        status = mpeg_time_seek_starttime(fp, indxfp, stime,
                                          &mpegtval,
                                          argvflags.has_xing);
        if (status == -1) {
            fprintf(stderr, "Invalid start time 2\n");
            return 1;
        }
        file_position += status;
        /*
         * Open external player program when -p is specified.  Write edit
         * data there when specified.  This preview mode allows for accurate
         * specification of edit points, by listening to the file at various
         * sample edit locations.
         */
        if (argvflags.p_flag) {
            if (!playfp) {
                playfp = fp_mpgdecoder_open(ttyio.mpctx, header.sample_rate,
                                            header.channel_mode != 3);
            }
            if (!playfp) {
                perror("popen");
                return 1;
            }
        }
        mpegfio_edit_segment(fp,
                             editfp,
                             &playfp,
                             &mpegtval,
                             &argvflags.edits[editsindx].etime,
                             &stats,
                             &file_position,
                             &ttyio,
                             argvflags.v_flag);


        /*
         * Display statistics with -s when there is no -p or -e flags.
         * There is no reason to be silent when no -e or -p is present.
         */
        if (!argvflags.s_flag ||
            (argvflags.s_flag && !argvflags.e_flag && !argvflags.p_flag)) {
            mpeg_file_stats_print(&stats, &mpegtval);
        }

        if (argvflags.has_xing) {
            if (editfp) {
                if (!argvflags.X_flag ||
                    (argvflags.X_flag && argvflags.X_val != 0))
                {
                    /*
                     * Update VBR header in edited output file.  Update frame 
                     * count and file size.
                     */
                    if (do_append) {
                        edit_xing_header(xingh.frames + stats.num_frames,
                                         xingh.bytes  + stats.file_size,
                                         xingh.xingbuf, 
                                         header.mpeg_version,
                                         header.channel_mode);
                    }
                    else {
                        edit_xing_header(stats.num_frames,
                                         stats.file_size, xingh.xingbuf,
                                         header.mpeg_version,
                                         header.channel_mode);
                    }
                    fclose(editfp);
                    editfp = fopen(edit_filename, "rb+");
                    if (editfp) {
                        fwrite(xingh.xingbuf, argvflags.has_xing, 1, editfp);
                        /*
                         * This fflush should not be needed here.  Testing on
                         * Linux, with glibc-2.1.3-15, this write is not
                         * always being flushed by the fclose.  Must be a
                         * glibc bug.
                         */
                        fflush(editfp);
                    }
                }
            }
            else if (argvflags.X_val == 1) {
                xingheader_edit(&xingh, stats.num_frames,
                                stats.file_size,
                                header.mpeg_version, header.channel_mode);
                fclose(fp);
                fp = fopen(argvflags.edits[editsindx].filename, "rb+");
                if (fp) {
                    fwrite(xingh.xingbuf, argvflags.has_xing, 1, fp);
                }
            }
        }

        if (indx_filename) {
            free(indx_filename);
            indx_filename = NULL;
        }

        if (fp) {
            fclose(fp);
            fp = NULL;
        }

        if (indxfp) {
            fclose(indxfp);
            indxfp = NULL;
        }
        if (!argvflags.out_fileappend && editfp) {
            fclose(editfp);
            editfp = NULL;
        }

        editsindx++;
        if (argvflags.out_filename) {
            argvflags.out_fileappend = 1;
        }
    } while (editsindx < argvflags.edits_cnt);
    fp_mpgdecoder_close(ttyio.mpctx);
    fp_mpgdecoder_free(ttyio.mpctx);

    if (editfp) {
        fclose(editfp);
        editfp = NULL;
    }

    return 0;
}
